@@ -5,10 +5,9 @@ import opcode
 import os
 import threading
 from collections import defaultdict
+from itertools import islice
 from os import getpid
 
-from colorama import AnsiToWin32
-
 from . import config
 from .util import BUILTIN_SYMBOLS
 from .util import CALL_COLORS
@@ -18,8 +17,11 @@ from .util import OTHER_COLORS
 from .util import PY3
 from .util import StringType
 from .util import builtins
+from .util import frame_iterator
+from .util import get_arguments
 from .util import iter_symbols
 from .util import safe_repr
+from colorama import AnsiToWin32
 
 try:
     from threading import get_ident
@@ -243,12 +245,12 @@ class ColorStreamAction(Action):
         """
         source = event.fullsource if full else event.source
         if source.startswith('??? NO SOURCE: '):
-            return '{SOURCE-FAILURE}??? NO SOURCE: {SOURCE-DETAIL}{}'.format(source[15:], **self.other_colors),
+            return '{SOURCE-FAILURE}??? NO SOURCE: {SOURCE-DETAIL}{}'.format(source[15:], **self.other_colors)
         elif source:
             return source
         else:
-            return '{SOURCE-FAILURE}??? NO SOURCE: {SOURCE-DETAIL}Source code string for module {!r} is empty.'.format(
-                event.module, **self.other_colors)
+            return '{SOURCE-FAILURE}??? NO SOURCE: {SOURCE-DETAIL}Source code string for {!r} is empty.'.format(
+                event.filename, **self.other_colors)
 
     def filename_prefix(self, event=None):
         """
@@ -257,11 +259,15 @@ class ColorStreamAction(Action):
         Returns: string
         """
         if event:
-            filename = event.filename or '<???>'
+            filename = event.filename
+            if filename == '?':
+                filename = '{SOURCE-FAILURE}?'.format(**self.other_colors)
+            lineno = '{COLON}:{LINENO}{:<5}'.format(event.lineno, **self.other_colors)
+
             if len(filename) > self.filename_alignment:
                 filename = '[...]{}'.format(filename[5 - self.filename_alignment:])
-            return '{:>{}}{COLON}:{LINENO}{:<5} '.format(
-                filename, self.filename_alignment, event.lineno, **self.other_colors)
+            return '{:>{}}{} '.format(
+                filename, self.filename_alignment, lineno, **self.other_colors)
         else:
             return '{:>{}}       '.format('', self.filename_alignment)
 
@@ -333,6 +339,9 @@ class ColorStreamAction(Action):
             **dict(self.other_colors, **kwargs)
         ))
 
+    def cleanup(self):
+        pass
+
 
 class CodePrinter(ColorStreamAction):
     """
@@ -418,9 +427,14 @@ class CallPrinter(CodePrinter):
     """
     EVENT_COLORS = CALL_COLORS
 
+    locals = defaultdict(list)
+
+    @classmethod
+    def cleanup(cls):
+        cls.locals = defaultdict(list)
+
     def __init__(self, *args, **kwargs):
         super(CallPrinter, self).__init__(*args, **kwargs)
-        self.locals = defaultdict(list)
 
     def __call__(self, event):
         """
@@ -437,50 +451,93 @@ class CallPrinter(CodePrinter):
         filename_prefix = self.filename_prefix(event)
 
         if event.kind == 'call':
-            code = event.code
-            stack.append(ident)
-            self.output(
-                '{}{}{}{KIND}{:9} {}{COLOR}=>{NORMAL} {}({}{COLOR}{NORMAL}){RESET}\n',
-                pid_prefix,
-                thread_prefix,
-                filename_prefix,
-                event.kind,
-                '   ' * (len(stack) - 1),
-                event.function,
-                ', '.join('{VARS}{VARS-NAME}{0}{VARS}={RESET}{1}'.format(
-                    var,
-                    self.try_str(event.locals.get(var, MISSING)) if event.detached else self.try_repr(event.locals.get(var, MISSING)),
-                    **self.other_colors
-                ) for var in code.co_varnames[:code.co_argcount]),
-                COLOR=self.event_colors.get(event.kind),
-            )
+            if event.builtin:
+                self.output(
+                    '{}{}{}{KIND}{:9} {}{COLOR} >{BUILTIN} {}.{}: {}{RESET}\n',
+                    pid_prefix,
+                    thread_prefix,
+                    filename_prefix,
+                    event.kind,
+                    '   ' * (len(stack) - 1),
+                    event.module,
+                    event.function,
+                    self.try_source(event).strip(),
+                    COLOR=self.event_colors.get(event.kind),
+                )
+            else:
+                code = event.code
+                stack.append(ident)
+                self.output(
+                    '{}{}{}{KIND}{:9} {}{COLOR}=>{NORMAL} {}({}{COLOR}{NORMAL}){RESET}\n',
+                    pid_prefix,
+                    thread_prefix,
+                    filename_prefix,
+                    event.kind,
+                    '   ' * (len(stack) - 1),
+                    event.function,
+                    ', '.join('{VARS}{0}{VARS-NAME}{1}{VARS}={RESET}{2}'.format(
+                        prefix,
+                        var_display,
+                        self.try_str(event.locals.get(var_lookup, MISSING))
+                        if event.detached
+                        else self.try_repr(event.locals.get(var_lookup, MISSING)),
+                        **self.other_colors
+                    ) for prefix, var_lookup, var_display in get_arguments(code)),
+                    COLOR=self.event_colors.get(event.kind),
+                )
         elif event.kind == 'exception':
-            self.output(
-                '{}{}{}{KIND}{:9} {}{COLOR} !{NORMAL} {}: {RESET}{}\n',
-                pid_prefix,
-                thread_prefix,
-                filename_prefix,
-                event.kind,
-                '   ' * (len(stack) - 1),
-                event.function,
-                self.try_str(event.arg) if event.detached else self.try_repr(event.arg),
-                COLOR=self.event_colors.get(event.kind),
-            )
-
+            if event.builtin:
+                self.output(
+                    '{}{}{}{KIND}{:9} {}{COLOR} ! {BUILTIN}{}.{}{RESET}\n',
+                    pid_prefix,
+                    thread_prefix,
+                    filename_prefix,
+                    event.kind,
+                    '   ' * (len(stack) - 1),
+                    event.module,
+                    event.function,
+                    COLOR=self.event_colors.get(event.kind),
+                )
+            else:
+                self.output(
+                    '{}{}{}{KIND}{:9} {}{COLOR} !{NORMAL} {}: {RESET}{}\n',
+                    pid_prefix,
+                    thread_prefix,
+                    filename_prefix,
+                    event.kind,
+                    '   ' * (len(stack) - 1),
+                    event.function,
+                    self.try_str(event.arg) if event.detached else self.try_repr(event.arg),
+                    COLOR=self.event_colors.get(event.kind),
+                )
         elif event.kind == 'return':
-            self.output(
-                '{}{}{}{KIND}{:9} {}{COLOR}<={NORMAL} {}: {RESET}{}\n',
-                pid_prefix,
-                thread_prefix,
-                filename_prefix,
-                event.kind,
-                '   ' * (len(stack) - 1),
-                event.function,
-                self.try_str(event.arg) if event.detached else self.try_repr(event.arg),
-                COLOR=self.event_colors.get(event.kind),
-            )
-            if stack and stack[-1] == ident:
-                stack.pop()
+            if event.builtin:
+                self.output(
+                    '{}{}{}{KIND}{:9} {}{COLOR} <{BUILTIN} {}.{}{RESET}\n',
+                    pid_prefix,
+                    thread_prefix,
+                    filename_prefix,
+                    event.kind,
+                    '   ' * (len(stack) - 1),
+                    event.module,
+                    event.function,
+                    COLOR=self.event_colors.get(event.kind),
+                )
+            else:
+                self.output(
+                    '{}{}{}{KIND}{:9} {}{COLOR}<={NORMAL} {}{}{RESET}{}\n',
+                    pid_prefix,
+                    thread_prefix,
+                    filename_prefix,
+                    event.kind,
+                    '   ' * (len(stack) - 1),
+                    event.function,
+                    '' if event.builtin else ': ',
+                    '' if event.builtin else (self.try_str(event.arg) if event.detached else self.try_repr(event.arg)),
+                    COLOR=self.event_colors.get(event.kind),
+                )
+                if stack and stack[-1] == ident:
+                    stack.pop()
         else:
             self.output(
                 '{}{}{}{KIND}{:9} {RESET}{}{}{RESET}\n',
@@ -535,7 +592,7 @@ class VarsPrinter(ColorStreamAction):
         for code, symbols in self.names.items():
             try:
                 obj = eval(code, dict(vars(builtins), **event.globals), event.locals)
-            except AttributeError:
+            except (AttributeError, KeyError):
                 continue
             except Exception as exc:
                 printout = '{INTERNAL-FAILURE}FAILED EVAL: {INTERNAL-DETAIL}{!r}'.format(exc, **self.other_colors)
@@ -656,10 +713,13 @@ class VarsSnooper(ColorStreamAction):
                         current_repr,
                     )
         if event.kind == 'return':
-            del self.stored_reprs[scope_key]
+            if scope_key in self.stored_reprs:
+                del self.stored_reprs[scope_key]
 
 
 RETURN_VALUE = opcode.opmap['RETURN_VALUE']
+if not PY3:
+    RETURN_VALUE = chr(RETURN_VALUE)
 
 
 class ErrorSnooper(CodePrinter):
@@ -728,25 +788,80 @@ class ErrorSnooper(CodePrinter):
                     detached_event = event.detach(self.try_repr)
                 self.events.append(detached_event)
                 if event.depth == self.origin.depth - 1:  # stop if the same function returned (depth is -1)
-                    if (event.code.co_code[event.frame.f_lasti] if PY3 else ord(event.code.co_code[event.frame.f_lasti])) == RETURN_VALUE:
+                    if event.instruction == RETURN_VALUE:
                         self.dump_events()
-                        self.output("{BRIGHT}{fore(BLACK)}{} function exit{RESET}\n", "-" * 46)
+                        self.output('{BRIGHT}{fore(BLACK)}{} function exit{RESET}\n', '-' * 46)
                     self.origin = None
                     self.events = None
             elif event.depth > self.origin.depth + self.max_depth:  # too many details
                 return
             elif len(self.events) > self.max_events:
                 self.dump_events()
-                self.output("{BRIGHT}{fore(BLACK)}{} too many lines{RESET}\n", "-" * 46)
+                self.output('{BRIGHT}{fore(BLACK)}{} too many lines{RESET}\n', '-' * 46)
             else:
                 if detached_event is None:
                     detached_event = event.detach(self.try_repr)
                 self.events.append(detached_event)
 
     def dump_events(self):
-        self.output("{BRIGHT}{fore(BLUE)}{} tracing {fore(YELLOW)}{}{fore(BLUE)} on {fore(RED)}{}{RESET}\n",
-                    ">" * 46, self.origin.function, self.origin.arg)
+        self.output('{BRIGHT}{fore(BLUE)}{} tracing {fore(YELLOW)}{}{fore(BLUE)} on {fore(RED)}{}{RESET}\n',
+                    '>' * 46, self.origin.function, self.origin.arg)
         for event in self.events:
             super(ErrorSnooper, self).__call__(event)
         self.origin = None
         self.events = None
+
+
+class StackPrinter(ColorStreamAction):
+    """
+    An action that prints a one-line stacktrace.
+
+    Args:
+        depth (int): The maximum number of frames to show.
+        limit (int): The maximum number of components to show in path. Eg: ``limit=2`` means it will show 1 parent: ``foo/bar.py``.
+        stream (file-like): Stream to write to. Default: ``sys.stderr``.
+        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
+        force_colors (bool): Force coloring. Default: ``False``.
+        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
+        repr_func (string or callable): Function to use instead of ``repr``.
+            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.
+    """
+
+    def __init__(self, depth=15, limit=2, **options):
+        self.limit = limit
+        self.depth = depth
+        super(StackPrinter, self).__init__(**options)
+
+    def __call__(self, event):
+        """
+        Handle event and print the stack.
+        """
+        pid_prefix = self.pid_prefix()
+        thread_prefix = self.thread_prefix(event)
+        filename_prefix = self.filename_prefix(event).rstrip()
+        sep = os.path.sep
+
+        if event.frame and event.frame.f_back:
+            template = '{}{}{}{CONT}:{BRIGHT}{fore(BLUE)}%s {KIND}<={RESET} %s' % (
+                event.function,
+                ' {KIND}<={RESET} '.join(
+                    '%s{CONT}:{RESET}%s{CONT}:{BRIGHT}{fore(BLUE)}%s' % (
+                        '/'.join(frame.f_code.co_filename.split(sep)[-self.limit:]),
+                        frame.f_lineno,
+                        frame.f_code.co_name
+                    )
+                    for frame in islice(frame_iterator(event.frame.f_back), self.depth)
+                )
+            )
+        else:
+            template = '{}{}{}{CONT}:{BRIGHT}{fore(BLUE)}%s {KIND}<= {BRIGHT}{fore(YELLOW)}no frames available {NORMAL}(detached=%s)' % (
+                event.function,
+                event.detached,
+            )
+        template += '{RESET}\n'
+        self.output(
+            template,
+            pid_prefix,
+            thread_prefix,
+            filename_prefix,
+        )
